<?php

/**
 * @author Tudor Ilisoi <tudorilisoi@yahoo.com>
 */
class XRelationBehavior extends CActiveRecordBehavior {

    static $supported_relations = array(
            CActiveRecord::HAS_MANY => 'HM_',
            CActiveRecord::HAS_ONE => 'HO_',
            CActiveRecord::MANY_MANY => 'MM_',
    );
    public $transactional = true;
    private $transaction;
    public $config = array();
    private $parsed_info = null;
    private $related_records = array();
    private $conf_defaults = array(
            'validate' => true,
            'stop_on_first_error' => true,
            'save' => true,
            'delete' => true,
    );
    public static function l($str) {
        $dbg = debug_backtrace();
        $offset = 1;
        $cls = isset($dbg[$offset]['class']) ? $dbg[$offset]['class'] : '?';
        $info = $cls . "::" . ((isset($dbg[$offset]['function'])) ? $dbg[$offset]['function'] : '');
        Yii::log("$info $str", CLogger::LEVEL_WARNING, 'X');
    }
    public function logID() {
        return get_class($this).'['.$this->owner->getPrimaryKey().']';
    }
    public function beforeValidate($event) {
        $this->process('beforeValidate', $event);
        return parent::beforeValidate($event);
    }

    public function beforeSave($event) {
        $this->process('beforeSave', $event);
        return parent::beforeSave($event);
    }

    public function afterSave($event) {
        $this->process('afterSave', $event);
        return parent::afterSave($event);
    }

    public function beforeDelete($event) {
        $this->process('beforeDelete', $event);
        return parent::afterSave($event);
    }

    public function parseInfo() {
        if ($this->parsed_info)
            return;
        $relations = $this->getOwner()->relations();
        foreach ($this->config as $rel => $conf) {
            $relation = $relations[$rel];
            /*
             * 0=type
             * 1=model class
             * 2=keys and config
            */
//                var_dump($relation);
            if (in_array($relation[0], array_keys(self::$supported_relations))) {
                $info = new stdClass;
                $info->type = $relation[0];
                $info->model_class = $relation[1];
                $info->keys = $relation[2];
                $info->prefix = self::$supported_relations[$info->type];
                $this->setParsedInfo($rel, $info);
                $this->related_records[$rel] = array();
                $specific_method = $info->prefix . 'ParseInfo';
//                        now call a method to populate relation specific info
                call_user_func(array($this, $specific_method), $rel);
            } else {
                throw new CException(get_class($this->getOwner()) . ': relation ' . $relation[0] . ' not supported');
            }
        }
    }

    /**
     *
     * @param string $event_name
     */
    public function process($event_name, $event) {
        $this->parseInfo();
        $short_name = strtolower(strtr($event_name, array('before' => '', 'after' => '')));
        if (($ret = $this->findByConfig($short_name))) {
            foreach ($ret as $rel) {
                $info = $this->getParsedInfo($rel);
                $method = $info->prefix . ucFirst($event_name);
                if (method_exists($this, $method)) {
                    self::l("$method on " . $this->getOwner()->logID() . "->" . $rel);
                    call_user_func_array(array($this, $method), array($rel, $event));
                } else {
                    self::l($this->getOwner()->logID() . "->" . $rel . ": $method not found, nothing to do");
                }
            }
        }
    }

    private function getParsedInfo($rel) {
        if (isset($this->parsed_info[$rel])) {
            return $this->parsed_info[$rel];
        }
        return false;
    }

    private function setParsedInfo($rel, $info) {

        $this->parsed_info[$rel] = $info;
    }

    private function HO_ParseInfo($rel) {
        $info = $this->getParsedInfo($rel);        
        $this->setParsedInfo($rel, $info);
    }

    private function HO_AttachRecords($relation, $record) {
        $info = $this->getParsedInfo($relation);
        if (!is_object($record) || !$record instanceof $info->model_class) {
            $cls = $info->model_class;
            $record = $cls::model()->findByPk($record);
        }
        $this->related_records[$relation] = $record;
    }

    private function HM_BeforeDelete($rel, $event) {
        $info = $this->getParsedInfo($rel);
        $current = $this->getOwner()->getRelated($rel, true);
        if (count($current))
            foreach ($current as $record) {
                $record->delete();
            }
    }

    private function HM_BeforeValidate($rel, $event) {
        $model = $this->getOwner();
        $related = $this->getAttachedRecords($rel);
        $valid = true;
        if (count($related)) {
            foreach ($related as $record) {
                $this->setChildFk($record, $rel);
//                self::l("validating HM $rel child " . $record->logID() . " for " . $this->getOwner()->logID());
                if (!$record->validate()) {
//                     $model->addErrors($record->getErrors());
//                    var_dump($record->getErrors());
                    $valid = false;
                }
            }
        }
        $event->isValid = $valid;
    }

    private function HM_afterSave($rel, $event) {
        $related = $this->getAttachedRecords($rel);
        if (count($related)) {
            foreach ($related as $record) {
                $this->setChildFk($record, $rel);
                self::l("saving HM $rel child for [" . $this->getOwner()->id . "]");
                $record->save();
            }
        }
    }

    private function HM_ParseInfo($rel) {
        $info = $this->getParsedInfo($rel);
        $info->fk = $info->keys;
        $this->setParsedInfo($rel, $info);
    }

    private function HM_AttachRecords($relation, $records) {
        $info = $this->getParsedInfo($relation);
//        var_dump($info);
        $conf = $this->config[$relation];
        try {
            foreach ($records as $record) {

                if(is_object($record) && !$record instanceof $info->model_class) {
                    throw new CException("Expecting $info->model_class, got [".get_class($record)."]" );
                }else
                if (is_scalar($record)) {
                    $cls = $info->model_class;
                    $record = $cls::model()->findByPk($record);
                    if (!$record)throw new CException("Attempt to attach non-existing $info->class_name");
                }

                if (isset($config->set_attributes)) {
                    foreach ($config->set_attributes as $attr => $value) {
                        $record->setAttribute($attr, $value);
                    }
                }
                $this->setChildFk($record, $relation);
                $this->related_records[$relation][] = $record;
            }
        } catch (Exception $e) {
            throw new CException($e->getMessage());
        }
    }

    private function MM_BeforeDelete($rel, $event) {
        $info = $this->getParsedInfo($rel);
        $obj = new $info->model_class;
        $table = $obj->getMetaData()->tableSchema;
        $fk = $table->primaryKey;
        $current = $this->getOwner()->getRelated($rel, true);
        $curr_keys = (array) $this->keysFromRecords($current, $fk);
        if (count($curr_keys))
            $this->MM_DeleteRecords($rel, $curr_keys);
//        Yii::app()->end();
    }

    private function MM_BeforeSave($rel, $event) {
//        var_dump($this->getParsedInfo($rel));        
    }

    private function MM_AfterSave($rel, $event) {
        $model = $this->getOwner();
        $info = $this->getParsedInfo($rel);
        $obj = new $info->model_class;
        $table = $obj->getMetaData()->tableSchema;
        $fk = $table->primaryKey;
        $related = $this->getAttachedRecords($rel);
        $current = $this->getOwner()->getRelated($rel, true);
        $old_keys = (array) $this->keysFromRecords($current, $fk);
        $curr_keys = (array) $this->keysFromRecords($related, $fk);
        $delete_keys = array_diff($old_keys, $curr_keys);
        $insert_keys = array_diff($curr_keys, $old_keys);
        if (count($delete_keys))
            $this->MM_DeleteRecords($rel, $delete_keys);
        if (count($insert_keys))
            $this->MM_InsertRecords($rel, $insert_keys);
    }

    private function MM_InsertRecords($rel, $keys) {
        $model = $this->getOwner();
        $pk = $model->getPrimaryKey();
        $info = $this->getParsedInfo($rel);
//        var_dump($info);
//        @TODO make room in config for item_class_id value and push it here
        $fnames = array($info->pk, $info->fk);
        foreach ($keys as $key) {
            $names = array();
            $values = array();
            $names[] = ':pk_' . $key;
            $names[] = ':fk_' . $key;
            $values[':pk_' . $key] = $pk;
            $values[':fk_' . $key] = $key;
            $sql = "INSERT IGNORE INTO " . $info->table . " (" . implode(', ', $fnames) . ") VALUES(" . implode("\n, ", $names) . ")";
            self::l($sql);
            $model->getCommandBuilder()->createSqlCommand($sql, $values)->execute();
        }
    }

    private function MM_DeleteRecords($rel, $keys) {
        $model = $this->getOwner();
        $pk = $model->getPrimaryKey();
        $info = $this->getParsedInfo($rel);
        $criteria = new CDbCriteria;
        $criteria->compare($info->pk, $pk, false);
//        XXX if() risky, deletes by pk only
        if ($keys)
            $criteria->addInCondition($info->fk, $keys, 'AND');
        //        @TODO make room in config for item_class_id value and push it here
        $model->getCommandBuilder()->createDeleteCommand($info->table, $criteria)->execute();
    }

    private function MM_ParseInfo($rel) {
        $info = $this->getParsedInfo($rel);
        self::l("Parsing info for " . $info->model_class . "->" . $rel);
        preg_match('/(\w+)\((\w+),(\s+)?(\w+)\)/', $info->keys, $matches);
//        var_dump($matches);
        $info->table = $matches[1];
        $info->pk = $matches[2];
        $info->fk = $matches[4];
        $this->setParsedInfo($rel, $info);
    }

    private function MM_AttachRecords($relation, $records) {
        $info = $this->getParsedInfo($relation);
        $conf = $this->config[$relation];
        try {
            foreach ($records as $record) {
                if (is_scalar($record) || !$record instanceof $info->model_class) {
                    $cls = $info->model_class;
                    $record = $cls::model()->findByPk($record);
                    if (!$record

                    )throw new CException("Attempt to attach non-existing $info->class_name");
                }

                if (isset($config->set_attributes)) {
                    foreach ($config->set_attributes as $attr => $value) {
                        $record->setAttribute($attr, $value);
                    }
                }

                $this->related_records[$relation][] = $record;
            }
        } catch (Exception $e) {
            throw new CException($e->getMessage());
        }
    }

    /**
     * attaches related records so XRelationBehavior can save and validate them using event hooks
     */
    public function attachRelatedRecords($data) {
        $this->parseInfo();
        foreach ($data as $relation => $records) {
            $info = $this->getParsedInfo($relation);
//            var_dump($this->parsed_info);
            call_user_func_array(array($this, $info->prefix . 'AttachRecords'), array($relation, $records));
            self::l("attached "
                    . count($this->related_records[$relation])
                    . " " . $info->model_class
                    . " records  to " . get_class($this->getOwner()));
        }
    }

    protected function getAttachedRecords($rel) {
        $this->parseInfo();
        return $this->related_records[$rel];
    }

    protected function keysFromRecords($records, $key) {
        $keys = null;
        foreach ($records as $r) {
            $keys[] = $r->getAttribute($key);
        }
        return $keys;
    }

    /**
     * sets related record foreign key to parent's foreign key
     * @param CactiveRecord $child_record
     * @param string $rel relation name
     */
    private function setChildFk($child_record, $rel) {
        $fk = $this->getChildFk($rel);
        $child_record->setAttribute($fk, $this->getOwner()->getPrimaryKey());
    }

    /**
     * get foreign key for a given relation name
     * @param string $rel relation name
     * @return string foreign key name
     */
    private function getChildFk($rel) {
        $relations = $this->getOwner()->relations();
//        XXX only works on non-composite foreign keys
        return $relations[$rel][2];
    }

    /**
     * finds configured relations for an event
     * @param string $conf_str one of validate, save, delete
     * @return array
     */
    private function findByConfig($conf_str) {
        $ret = array();
        foreach ($this->config as $rel => $conf) {
            $conf = array_merge($this->conf_defaults, $conf);
            if ($conf[$conf_str])
                $ret[] = $rel;
        }
        return $ret;
    }

    private function beginTransaction() {
        $model = $this->getOwner();
        if ($this->transactional && !$model->dbConnection->currentTransaction) {
            $this->transaction = $model->dbConnection->beginTransaction();
        }
    }

    private function endTransaction() {
        if ($this->transactional && $this->transaction)
            if (!$this->hasError) {
                $this->transaction->commit();
            } else {
                $this->transaction->rollBack();
            }
    }

}
